"""
Tests that sktran_do modules is operating as expected.

Date: March 28, 2017
Geometry: TEMPO at (20, -100, 35786000), looking at latitudes [-38.51, 26.05], longitudes [-100, -100]
"""

import unittest, math
import numpy as np
import sasktranif.sasktranif as skif

import sys
def printf(format, *args):
    sys.stdout.write(format % args)

class QuickDOTest(unittest.TestCase):

    def test_weighting_functions(self):

        #
        # In this example we will calculate the weighting functions for ozone in a simple atmosphere (rayleigh + ozone).
        # First we'll get the analytic weighting functions using the discrete-ordinates engine, then we'll compute the
        # finite-difference weighting functions by perturbing the number density of ozone at the interesting altitudes.
        #

        engine = skif.ISKEngine('DO')
        engine.SetProperty('numstreams', 16)
        engine.SetProperty('numlayers', 25)
        engine.SetProperty('losdiagnostics', 1)

        # First create the simple atmosphere
        msis90 = skif.ISKClimatology('MSIS90')
        labow_ozone = skif.ISKClimatology('O3LABOW')
        opt_rayleigh = skif.ISKOpticalProperty('RAYLEIGH')
        opt_ozone = skif.ISKOpticalProperty('O3_OSIRISRES')
        wavelengths = [310]

        engine.SetAtmosphericState(msis90)
        engine.AddSpecies(skif.SKCLIMATOLOGY_AIRNUMBERDENSITY_CM3, msis90, opt_rayleigh)
        engine.AddSpecies(skif.SKCLIMATOLOGY_O3_CM3, labow_ozone, opt_ozone)
        engine.SetWavelengths(wavelengths)

        # Configure surface
        lambertian_brdf = skif.ISKBrdf('Lambertian')
        lambertian_brdf.SetProperty('Albedo', 0.3)

        engine.SetBRDF(lambertian_brdf)

        # Configure geometry
        sun = [-0.190809, -0.98162718, 0]
        obs = [-6880579.70829706, -39021706.60809562, 14407230.65919554]
        look = [0.15042014, 0.87062868, -0.46837964]
        mjd = 54372.5

        engine.SetProperty('sun', sun)
        engine.AddLineOfSight(mjd, obs, look)

        # Configure weighting functions specifications:
        wf_altitudes = np.linspace(1e3, 51e3, 26)           # altitudes to calculate weighting functions at
        wf_widths = np.array([1000] * len(wf_altitudes))    # width of perturbation (above and below the given altitude)
        wf_eps = np.array([0.01] * len(wf_altitudes))       # relative change in species number density (perturbation)

        engine.SetProperty('wfclimatologies', 'SKCLIMATOLOGY_O3_CM3')
        engine.SetProperty('wfaltitudes', wf_altitudes)
        engine.SetProperty('wfwidths', wf_widths)
        engine.SetProperty('wfeps', wf_eps)
        engine.SetProperty('wfform', 0)                 # we want dI_dLogX


        engine.InitializeModel()
        ok, rad = engine.CalculateRadiance()
        ok, wf = engine.GetWeightingFunctions()

        wf_correct = np.array([
            # May 19, 2017
            -8.16129033e-05, -1.22650298e-04, -1.48115362e-04, -1.60420953e-04,
            -1.80710717e-04, -2.31164826e-04, -2.85136908e-04, -4.04224888e-04,
            -6.72436706e-04, -1.06773579e-03, -1.39997215e-03, -1.63143755e-03,
            -1.81549434e-03, -1.78586686e-03, -1.57352013e-03, -1.26769933e-03,
            -1.00312499e-03, -7.55823885e-04, -5.84603170e-04, -4.13461450e-04,
            -2.74406814e-04, -1.72550733e-04, -1.06845920e-04, -6.74153699e-05,
            -4.31475030e-05, -2.77138122e-05
        ])
        self.assertEqual(len(wf.flatten()), len(wf_correct))
        for w, c in zip(wf.flatten(), wf_correct):
            self.assertAlmostEqual(w, c)

        """
        #
        # Next we'll perturb the labow_ozone climatology and check that the analytic weighting functions are correct.
        #
        ok, refpt = engine.GetProperty('referencepoint[0]')

        ptrb_rad = [None] * 26
        clim_altitudes = np.linspace(0, 100e3, 201)
        for i in range(0, 26):
            ref_inst = skif.GEODETIC_INSTANT()
            ref_inst.FromSequence(refpt)

            ok, nd_o3 = labow_ozone.GetHeightProfile(skif.SKCLIMATOLOGY_O3_CM3, ref_inst, clim_altitudes)

            eps = 0.01
            ptrb_clim_index = 4 * i + 2
            nd_at_wf_altitude = nd_o3[ptrb_clim_index]

            nd_o3[ptrb_clim_index - 1]  += 0.5 * nd_at_wf_altitude * eps
            nd_o3[ptrb_clim_index]      += nd_at_wf_altitude * eps
            nd_o3[ptrb_clim_index + 1]  += 0.5 * nd_at_wf_altitude * eps

            ptrbed_ozone_clim = skif.ISKClimatology('USERDEFINED_PROFILE')
            ptrbed_ozone_clim.SetProperty('DoLogInterpolation', 0)
            ptrbed_ozone_clim.SetProperty('DoPieceWiseLinear', 1)
            ptrbed_ozone_clim.SetProperty('Heights', clim_altitudes)
            ptrbed_ozone_clim.SetPropertyUserDefined(skif.SKCLIMATOLOGY_O3_CM3, nd_o3)

            engine = skif.ISKEngine('DO')
            engine.SetProperty('numstreams', 16)
            engine.SetProperty('numlayers', 25)

            engine.SetAtmosphericState(msis90)
            engine.AddSpecies(skif.SKCLIMATOLOGY_AIRNUMBERDENSITY_CM3, msis90, opt_rayleigh)
            engine.AddSpecies(skif.SKCLIMATOLOGY_O3_CM3, ptrbed_ozone_clim, opt_ozone)     # perturbed ozone climatology
            engine.SetWavelengths(wavelengths)

            engine.SetBRDF(lambertian_brdf)
            engine.SetProperty('sun', sun)
            engine.AddLineOfSight(mjd, obs, look)

            ok, ptrb_rad[i] = engine.CalculateRadiance()

        fd = [None] * 26
        for i in range(0, 26):
            fd[i] = (ptrb_rad[i][0,0] - rad[0,0]) / 0.01

        # The following can be used to see a comparison of the two:
        # import matplotlib.pyplot as plt
        # plt.figure()
        # plt.plot(wf[0,0,:], wf_altitudes, fd, wf_altitudes)
        # plt.legend(['Analytic', 'Finite-Difference'])
        # plt.show()
        """

    def test_diagnostics(self):
        geo = skif.GEODETIC_INSTANT()
        geo.FromSequence([0, 0, 0, 54372])

        clim_air = skif.ISKClimatology('MSIS90')
        opt_air = skif.ISKOpticalProperty('RAYLEIGH')

        engine = skif.ISKEngine("DO")
        
        engine.SetProperty('numstreams', 40)
        engine.SetProperty('numlayers', 200)

        engine.AddSpecies(skif.SKCLIMATOLOGY_AIRNUMBERDENSITY_CM3, clim_air, opt_air)
        engine.AddSpecies(skif.SKCLIMATOLOGY_O3_CM3, skif.ISKClimatology("O3LABOW"),
                          skif.ISKOpticalProperty("O3_OSIRISRES"))

        engine.SetAtmosphericState(clim_air)

        sun = np.array([-0.190809, -0.98162718, 0])
        engine.SetProperty('sun', sun)

        obs = np.array([ -6880579.70829706, -39021706.60809562, 14407230.65919554])
        looks = np.array([[ 0.15042014, 0.87062868, -0.46837964],
                          [ 0.1518094, 0.87854755, -0.45288863]])
        mjd = 54372.5
        for los in np.reshape(looks, (-1, 3)):
            engine.AddLineOfSight(mjd, obs, los)

        engine.SetWavelengths([350, 400, 700])

        lambertian_brdf = skif.ISKBrdf('Lambertian')
        lambertian_brdf.SetProperty('Albedo', 0.3)

        engine.SetBRDF(lambertian_brdf)

        engine.SetProperty('diagnostics', 1.0)

        engine.InitializeModel()
        ok, rad = engine.CalculateRadiance()

        # Reconstruct the solutions
        for w in [0, 1, 2]: # wavelengths: 350, 400, 700
            for l in [0, 1]: # 2 look directions
                ok = True
                # Geometry
                ok, ref = engine.GetProperty('referencepoint[%i]' % l)
                ok, solar_zenith =  engine.GetProperty('localsolarzenith[%i]' % l)
                ok, los_zenith = engine.GetProperty('localloszenith[%i]' % l)
                ok, rel_azimuth = engine.GetProperty('localrelativeazimuth[%i]' % l)
                # Atmospheric properties
                num_los = 2
                wl_index = w * num_los + l
                ok, optical_depths = engine.GetProperty('daopticaldepths[%i, %i]' % (w, l))
                ok, single_scatter_albedo = engine.GetProperty('dassa[%i, %i]' % (w, l))
                ok, phasef_expansion = engine.GetProperty('daphasef[%i, %i]' % (w, l))
                # Radiative transfer solutions
                ok, radiance_components = engine.GetProperty('radiancecomponents[%i, %i]' % (w, l))
                ok, reflected_components = engine.GetProperty('reflectedcomponents[%i, %i]' % (w, l))
                ok, participating_source_terms = engine.GetProperty('participatingsourceterms[%i, %i]' % (w, l))
                # Calculate radiance
                radiance = 0
                for m in range(0, 40):
                    component = reflected_components[m] # radiance at the ground

                    for od, pst in zip(optical_depths[::-1], participating_source_terms[-40+m::-40]): # integrate up to TOA
                        component = component * math.exp(-od/math.cos(los_zenith)) + pst
                    component *= math.cos(m * rel_azimuth)
                    self.assertAlmostEqual(component, radiance_components[m])
                    radiance += component
                self.assertAlmostEqual(radiance, rad[w][l])

    def test_lambertian(self):
        geo = skif.GEODETIC_INSTANT()
        geo.FromSequence([0, 0, 0, 54372])

        clim_air = skif.ISKClimatology('MSIS90')
        opt_air = skif.ISKOpticalProperty('RAYLEIGH')

        engine = skif.ISKEngine("DO")
        engine.SetProperty('numstreams', 40)
        engine.SetProperty('numlayers', 200)

        engine.AddSpecies(skif.SKCLIMATOLOGY_AIRNUMBERDENSITY_CM3, clim_air, opt_air)
        engine.AddSpecies(skif.SKCLIMATOLOGY_O3_CM3, skif.ISKClimatology("O3LABOW"),
                          skif.ISKOpticalProperty("O3_OSIRISRES"))

        engine.SetAtmosphericState(clim_air)

        sun = np.array([-0.190809, -0.98162718, 0])
        engine.SetProperty('sun', sun)

        obs = np.array([ -6880579.70829706, -39021706.60809562, 14407230.65919554])
        looks = np.array([[ 0.15042014, 0.87062868, -0.46837964],
                          [ 0.1518094, 0.87854755, -0.45288863],
                          [ 0.15315242, 0.8861988, -0.43725967],
                          [ 0.15444878, 0.89358011, -0.42149752],
                          [ 0.1556981, 0.90068923, -0.40560698],
                          [ 0.15689999, 0.90752398, -0.38959288],
                          [0.15805408, 0.9140823, -0.37346012],
                          [0.15916003, 0.92036218, -0.35721359],
                          [0.1602175, 0.9263617, -0.34085825],
                          [0.16122617, 0.93207905, -0.32439908]])
        mjd = 54372.5
        for los in np.reshape(looks, (-1, 3)):
            engine.AddLineOfSight(mjd, obs, los)

        engine.SetWavelengths([350, 400, 700])

        lambertian_brdf = skif.ISKBrdf('Lambertian')
        lambertian_brdf.SetProperty('Albedo', 0.3)

        engine.SetBRDF(lambertian_brdf)

        engine.InitializeModel()
        ok, rad = engine.CalculateRadiance()

        correct = np.array([
            [ 0.14010926, 0.13446559, 0.13185127, 0.12991796, 0.12810804, 0.12622069
            , 0.12416023, 0.12186823, 0.11928992, 0.11640004],
            [ 0.11911075, 0.11578273, 0.11526143, 0.11484618, 0.11409809, 0.11291455
            , 0.11126909, 0.10915426, 0.10656346, 0.10348698],
            [ 0.07810069, 0.08722341, 0.09215029, 0.0947986 , 0.09591001, 0.09584553
            , 0.09478171, 0.09282681, 0.08998259, 0.0862821 ]
        ])

        self.assertEqual(len(rad.flatten()), len(correct.flatten()))
        for r, c in zip(rad.flatten(), correct.flatten()):
            self.assertAlmostEqual(r, c)

    def test_pineforest(self):
        geo = skif.GEODETIC_INSTANT()
        geo.FromSequence([0, 0, 0, 54372])

        clim_air = skif.ISKClimatology('MSIS90')
        opt_air = skif.ISKOpticalProperty('RAYLEIGH')

        engine = skif.ISKEngine("DO")
        engine.SetProperty('numstreams', 40)
        engine.SetProperty('numlayers', 200)

        engine.AddSpecies(skif.SKCLIMATOLOGY_AIRNUMBERDENSITY_CM3, clim_air, opt_air)
        engine.AddSpecies(skif.SKCLIMATOLOGY_O3_CM3, skif.ISKClimatology("O3LABOW"),
                          skif.ISKOpticalProperty("O3_OSIRISRES"))

        engine.SetAtmosphericState(clim_air)

        sun = np.array([-0.190809, -0.98162718, 0])
        engine.SetProperty('sun', sun)

        obs = np.array([ -6880579.70829706, -39021706.60809562, 14407230.65919554])
        looks = np.array([[ 0.15042014, 0.87062868, -0.46837964],
                          [ 0.1518094, 0.87854755, -0.45288863],
                          [ 0.15315242, 0.8861988, -0.43725967],
                          [ 0.15444878, 0.89358011, -0.42149752],
                          [ 0.1556981, 0.90068923, -0.40560698],
                          [ 0.15689999, 0.90752398, -0.38959288],
                          [0.15805408, 0.9140823, -0.37346012],
                          [0.15916003, 0.92036218, -0.35721359],
                          [0.1602175, 0.9263617, -0.34085825],
                          [0.16122617, 0.93207905, -0.32439908]])
        mjd = 54372.5
        for los in np.reshape(looks, (-1, 3)):
            engine.AddLineOfSight(mjd, obs, los)

        engine.SetWavelengths([350, 400, 700])

        roujean_brdf = skif.ISKBrdf('Roujean')
        roujean_brdf.SetProperty('SetPredefinedParameters', 8)

        engine.SetBRDF(roujean_brdf)

        engine.InitializeModel()
        ok, rad = engine.CalculateRadiance()

        correct = np.array([
            [ 0.11198657, 0.09518884, 0.08603374, 0.08006655, 0.07595647, 0.07311574
            , 0.07122159, 0.07007446, 0.0695374 , 0.06952421],
            [ 0.08177645, 0.06563946, 0.05805092, 0.05345452, 0.05043561, 0.04842605
            , 0.04713995, 0.04641629, 0.04616028, 0.04632018],
            [ 0.02297224, 0.01893762, 0.01716691, 0.01611499, 0.01542312, 0.01495173
            , 0.01463013, 0.01442208, 0.01429883, 0.014251  ]
        ])
        self.assertEqual(len(rad.flatten()), len(correct.flatten()))
        for r, c in zip(rad.flatten(), correct.flatten()):
            self.assertAlmostEqual(r, c)

if __name__ == "__main__":
    unittest.main()